# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
Module protocol provides static methods to access Target Communication
Framework root objects:

1. the framework event queue and dispatch thread
2. local instance of Locator service, which maintains a list of available
   targets
3. list of open communication channels.

It also provides utility methods for posting asynchronous events,
including delayed events (timers).
"""

import sys
import uuid
import pytcf
import atexit

_event_queue = None


def startEventQueue(logfile=""):
    global _event_queue
    pytcf.launch_event_loop(logfile)
    _event_queue = True



@atexit.register
def stopEventQueue():
    global _event_queue
    pytcf.stop_event_loop()
    _event_queue = False


def getEventQueue():
    """
    @return instance of event queue that is used for TCF events.
    """
    return _event_queue


def isDispatchThread():
    """
    Returns true if the calling thread is the TCF event dispatch thread.
    Use this call to ensure that a given task is being executed (or not being)
    on dispatch thread.

    @return true if running on the dispatch thread.
    """
    return pytcf.is_dispatch_thread()


def invokeLater(c, *args, **kwargs):
    """
    Causes callable to be called with given arguments
    in the dispatch thread of the framework.
    Events are dispatched in same order as queued.
    If invokeLater is called from the dispatching thread
    the callable will still be deferred until
    all pending events have been processed.

    This method can be invoked from any thread.

    @param c - the callable to be executed asynchronously
    """
    pytcf.post_event(c, *args, **kwargs)


def invokeLaterWithDelay(delay, c, *args, **kwargs):
    """
    Causes callable event to called in the dispatch thread of the framework.
    The event is dispatched after given delay.

    This method can be invoked from any thread.

    @param delay     milliseconds to delay event dispatch.
                     If delay <= 0 the event is posted into the
                     "ready" queue without delay.
    @param c - the callable to be executed asynchronously.
    """
    pytcf.post_event_with_delay(delay, c, *args, **kwargs)


def invokeAndWait(c, *args, **kwargs):
    """
    Causes callable to be called in the dispatch thread of the framework.
    Calling thread is suspended until the method is executed.
    If invokeAndWait is called from the dispatching thread
    the callable is executed immediately.

    This method can be invoked from any thread.

    @param c  the callable to be executed on dispatch thread.
    """
    return pytcf.post_event_and_wait(c, *args, **kwargs)


_agentID = str(uuid.uuid4())


def getAgentID():
    return _agentID


_logger = None


def setLogger(logger):
    """
    Set framework logger.
    By default sys.stderr is used.

    @param logger - an object implementing Logger interface.
    """
    global _logger
    _logger = logger


def log(msg, x=None):
    """
    Logs the given message.
    @see #setLogger
    This method can be invoked from any thread.
    @param msg - log entry text
    @param x - an exception associated with the log entry or None.
    """
    if not _logger:
        sys.stderr.write(str(msg) + '\n')
        while x:
            import traceback
            sys.stderr.write("%s: %s\n" % (type(x).__name__, x))
            tb = getattr(x, "tb", None) or sys.exc_info()[2]
            if tb:
                traceback.print_tb(tb)
            caused_by = getattr(x, "caused_by", None)
            if caused_by:
                sys.stderr.write("Caused by: " + str(caused_by) + "\n")
                x = caused_by
            else:
                break
    else:
        _logger.log(msg, x)


def startDiscovery():
    pass


def shutdownDiscovery():
    pass


def getLocator():
    """
    Get instance of the framework locator service.
    The service can be used to discover available remote peers.
    @return instance of LocatorService.
    """
    from ..services.local.LocatorService import LocatorService
    return LocatorService.locator


def getOpenChannels():
    """
    Return an array of all open channels.
    @return an array of IChannel
    """
    return pytcf.get_channels()


class ChannelOpenListener(object):
    """
    Interface to be implemented by clients willing to be notified when
    new TCF communication channel is opened.

    The interface allows a client to get pointers to channel objects that were
    opened by somebody else. If a client open a channel itself, it already has
    the pointer and does not need protocol.ChannelOpenListener. If a channel is
    created, for example, by remote peer connecting to the client, the only way
    to get the pointer is protocol.ChannelOpenListener.
    """
    def __init__(self):
        self.active = True

    def onChannelOpen(self, channel):
        pass

    def __call__(self, channel):
        if self.active:
            self.onChannelOpen(channel)


def addChannelOpenListener(listener):
    """
    Add a listener that will be notified when new channel is opened.
    @param listener
    """
    assert isDispatchThread()
    pytcf.add_channel_open_listener(listener)


def removeChannelOpenListener(listener):
    """
    Remove channel opening listener.
    @param listener
    """
    assert isDispatchThread()
    listener.active = False


def sendEvent(service_name, event_name, data):
    """
    Transmit TCF event message.
    The message is sent to all open communication channels - broadcasted.
    """
    assert isDispatchThread()
    pytcf.write_event_args(service_name, event_name, data)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from collections import deque
import argparse
import sys
import os
from pathlib import Path
from typing import Tuple, Dict, Union

from more_itertools import one

import chipscopy
from chipscopy import create_session, report_versions, report_devices
from chipscopy.api.session import Session
from chipscopy.client.view_info import ViewInfo
from chipscopy.dm import Node


class NodeVisitorBase:
    """
    In ChipScoPy, nodes are organized as a tree under a view.
    Example hw_server views are: jtag, memory, debugcore
    Example cs_server views are: chipscope

    Hierarchy:
    server -> view -> node tree

    NodeVisitorBase is a base class that makes visiting nodes easy. Just make a class with the
    NodeVisitorBase as a parent, and define functions with visit_<cls_name> for the node types
    you want to visit. <cls_name> represents the node class type like "JtagDevice", etc.
    """

    def __init__(self, view):
        self._view = view

    def walk(self, node_list=None):
        """
        Recursively walk all nodes in view, including this node
        """
        if node_list is None:
            nodes_to_walk = list(self._view.get_children())
        else:
            nodes_to_walk = list(node_list)

        if len(nodes_to_walk) > 0:
            last_node = nodes_to_walk[-1]
        else:
            last_node = True
        if nodes_to_walk:
            for current_node in nodes_to_walk:
                self._visit(current_node, current_node == last_node)

    def _visit(self, node, is_last_sibling):
        cls = node.props.get("node_cls")
        # Example: <class \'chipscopy.client.jtagdevice.JtagCable\'>
        match = re.fullmatch("^<class \\'(.+)\\'>$", str(cls))
        full_class_name = match.group(1)
        class_name = full_class_name.split(".")[-1]
        method = "_visit_" + class_name
        visitor_method = getattr(self, method, self._visit_default)
        upgraded_node = self._view.get_node(node.ctx, cls)
        assert upgraded_node is not None
        return visitor_method(upgraded_node, is_last_sibling)  # noqa

    def _visit_default(self, node, is_last_sibling):
        children = list(self._view.get_children(node))
        self.walk(children)

    @staticmethod
    def _print_node_path(node_name, parent_names):
        parent_names = deque(parent_names)
        while parent_names:
            name = parent_names.popleft()
            print(f"{name}/", end="")
        print(f"{node_name}")


# GLOBALS FOR COMMAND LINE MODULE
_hw_url: str = os.getenv("HW_SERVER_URL", "localhost:3121")
_cs_url: str = os.getenv("CS_SERVER_URL", "localhost:3042")
_bypass_version_check: bool = True


def display_banner():
    print(f"\n******** Xilinx ChipScoPy v{chipscopy.__version__}")
    print("  ****** Copyright (C) 2021-2022 Xilinx, Inc. All Rights Reserved.\n")
    print("  ****** Copyright (C) 2022-2023 Advanced Micro Devices, Inc. All Rights Reserved.\n")
    print("WARNING: Commands and options are subject to change.")
    print()


def report(session: Session, *, devices: bool = True, servers: bool = True):
    if servers:
        report_versions(session)
    if devices:
        report_devices(session)


def _get_node_path(host: str, view: ViewInfo, view_name: str, node: Node) -> str:
    name = node.props.get("Name", "")
    path = ""
    if name:
        hier_name = ""
        parent_ctx = node.parent_ctx
        while parent_ctx:
            parent = view.get_node(parent_ctx)
            hier_name = f"{hier_name}{parent.props.get('Name')}/"
            parent_ctx = parent.parent_ctx
        path = f"{host}/{view_name}/{hier_name}{name}"
    return path


def _get_all_nodes(session: Session) -> Dict[Tuple[str, str], Tuple[int, str, ViewInfo, Node]]:
    retval = {}
    views = ["jtag", "memory", "debugcore", "chipscope"]
    index = 0
    for view_name in views:
        if view_name == "chipscope" and session.cs_server:
            host = session.cs_server.props.get("url")
            view = session.cs_server.get_view(view_name)
            if not view:
                # Not an error for no cs_server - just ignore
                continue
        elif session.hw_server:
            host = session.hw_server.props.get("url")
            view = session.hw_server.get_view(view_name)
        else:
            raise RuntimeError(f"Unable to get view {view_name} from server")
        # Do a depth first search - neater results for numbering (don't use view.get_all())
        node_list = deque(view.get_children())
        while node_list:
            node = node_list.popleft()
            if node.props.get("Name"):
                retval[(view_name, node.ctx)] = (index, host, view, node)
                index += 1
            children = list(view.get_children(node))
            children.reverse()
            node_list.extendleft(children)
    return retval


def program(
    session: Session,
    *,
    file: Union[str, Path] = "",
    dna: str = "",
    cable_context: str = "",
    context: str = "",
    part: str = "",
    family: str = "",
    jtag_index: str = "",
    device_index: str = "",
    skip_reset: bool = False,
    list_: bool = False,
    program_log: bool = False,
    force_reset: bool = False,
):
    def print_device(device_):
        d = device_.to_dict()
        keys = ["jtag_index", "part", "dna", "context", "cable_context"]
        for key in keys:
            if key == "dna" and d.get(key):
                val = hex(d.get(key))
            else:
                val = d.get(key)
            print(f"  {key}:{val}", end="")
        print()

    all_devices = session.devices

    args = {}

    if dna:
        args["dna"] = int(dna, 0)
    if context:
        args["context"] = context
    if cable_context:
        args["cable_context"] = cable_context
    if context:
        args["context"] = context
    if part:
        args["part"] = part
    if family:
        args["family"] = family
    if jtag_index:
        args["jtag_index"] = int(jtag_index)

    if list_:
        if len(args) > 0:
            raise ValueError("--list is not allowed with other options")
        for idx, device in enumerate(all_devices):
            print(f"DEVICE #{idx}:", end="")
            print_device(device)
        return

    if len(all_devices) == 0:
        raise IndexError("No devices in device list")

    if not file and not program_log and not force_reset:
        raise RuntimeError("No programming file specified.")

    if device_index:
        device_index = int(device_index)
        if len(args) > 0:
            raise ValueError("Device index is not allowed with other options")
        if len(all_devices) <= device_index:
            raise ValueError("Device index out of range")
        device = all_devices[device_index]
    else:
        if len(args) > 0:
            device = all_devices.get(**args)
        elif len(all_devices) == 2 and all_devices[0].to_dict().get("part") == "arm_dap":
            # Ok to ignore arm_dap on same versal
            device = all_devices[1]
        else:
            if len(all_devices) > 1:
                raise ValueError("Multiple devices match selection")
            device = one(all_devices)

    if force_reset:
        print("Resetting Device", end="")
        print_device(device)
        device.reset()
        print()

    if file:
        print("Programming:", end="")
        print_device(device)
        device.program(file, skip_reset=skip_reset)
        print()

    if program_log:
        print()
        print(device.get_program_log())
        print()


class TreePrinter(NodeVisitorBase):
    tree_glyphs_std = {"blank": "    ", "line": "│   ", "middle": "├── ", "last": "└── "}
    tree_glyphs_alt = {"blank": "    ", "line": "|   ", "middle": "|-- ", "last": "\\-- "}
    tree_glyphs_space = {"blank": "    ", "line": "    ", "middle": "    ", "last": "    "}
    tree_glyphs_none = {"blank": "", "line": "", "middle": "", "last": ""}
    tree_glyphs_selector = {
        "std": tree_glyphs_std,
        "alt": tree_glyphs_alt,
        "space": tree_glyphs_space,
        "none": tree_glyphs_none,
    }

    def __init__(self, session, view_name, print_properties):
        self._tree_glyphs = TreePrinter.tree_glyphs_std
        self._show_context = True
        self._jtag_index = None
        self.session = session
        self._view_name = view_name
        self._indent_prefix = list()
        self._print_properties = print_properties
        self._all_nodes = _get_all_nodes(self.session)

        if view_name == "chipscope":
            view = self.session.cs_server.get_view(view_name)
        else:
            view = self.session.hw_server.get_view(view_name)
        super().__init__(view)

    def print(self, *, show_context=False, glyph_type="std"):
        self._tree_glyphs = TreePrinter.tree_glyphs_selector.get(glyph_type)
        self._jtag_index = None
        self._show_context = show_context
        print(self._view_name)
        self._indent_prefix = list()
        self.walk()

    def _print_node_properties(self, node: Node, node_str: str, prefix: str):
        props_dict = node.props
        prop_prefix = " " * len(prefix)
        for key, val in props_dict.items():
            if key != "regs":
                print(f"{prop_prefix}    {key:}: {val}")
        # Special printing for all regs - they need to be refreshed
        regs = node.props.get("regs")
        if regs:
            reg_names_to_refresh = tuple(regs.keys())
            node.update_regs(reg_names=reg_names_to_refresh, force=True, done=None)
            for reg_key, reg_val in regs.items():
                reg_data_len = len(reg_val.data)
                for index in range(0, reg_data_len):
                    int_value = int.from_bytes(
                        reg_val.data[index], byteorder="little", signed=False
                    )
                    if reg_data_len == 1:
                        print(f"{prop_prefix}    regs.{reg_key:}: {hex(int_value)}")
                    else:
                        print(f"{prop_prefix}    regs.{reg_key:}.slr{index}: {hex(int_value)}")

    def _print_node(self, node: Node, prefix: str):
        name = node.props.get("Name")
        cls = node.props.get("node_cls")
        # Example: <class \'chipscopy.client.jtagdevice.JtagCable\'>
        match = re.fullmatch("^<class \\'(.+)\\'>$", str(cls))
        full_class_name = match.group(1)
        class_name = full_class_name.split(".")[-1]

        found_val = self._all_nodes.get((self._view_name, node.ctx))

        try:
            # index:int, ctx:str, view:ViewInfo, node:Node
            (index, _, _, _) = self._all_nodes[(self._view_name, node.ctx)]
        except KeyError as ex:
            index = "?"
        node_str = f"{prefix}{index}- {name} ({class_name})"
        if self._show_context:
            indent_character_count = 50 - len(node_str)
            if indent_character_count < 0:
                indent_character_count = 2
            indent = " " * indent_character_count
            print(f"{node_str}{indent}{node.ctx}")
        else:
            print(f"{node_str}")
        if self._print_properties:
            self._print_node_properties(node, node_str, prefix)

    def _visit_JtagCable(self, node: Node, is_last_sibling: bool):
        # Tracking JtagCable and JtagDevice so we can keep running jtag_index per device in the chain
        self._jtag_index = 0
        self._visit_default(node, is_last_sibling)

    def _visit_JtagDevice(self, node: Node, is_last_sibling: bool):
        # Tracking JtagCable and JtagDevice so we can keep running jtag_index per device in the chain
        self._jtag_index += 1
        self._visit_default(node, is_last_sibling)

    def _visit_default(self, node: Node, is_last_sibling: bool):
        if is_last_sibling:
            glyph = self._tree_glyphs["last"]
        else:
            glyph = self._tree_glyphs["middle"]
        self._indent_prefix.append(glyph)
        prefix = "".join([i for i in self._indent_prefix])
        self._print_node(node, prefix)
        children = list(self._view.get_children(node))
        # Need to update prefix glyphs before moving to next nodes
        for idx, item in enumerate(self._indent_prefix):
            if item == self._tree_glyphs["last"]:
                self._indent_prefix[idx] = self._tree_glyphs["blank"]
            elif item == self._tree_glyphs["middle"]:
                self._indent_prefix[idx] = self._tree_glyphs["line"]
        self.walk(children)
        self._indent_prefix.pop()


def tree(
    session: Session, *, view=None, show_context=True, glyph_type="std", print_properties=False
):
    if isinstance(view, str):
        views = [view]
    elif isinstance(view, list):
        views = view
    else:
        views = ["jtag", "memory", "debugcore", "chipscope"]

    for view in views:
        tree_printer = TreePrinter(session, view, print_properties)
        tree_printer.print(show_context=show_context, glyph_type=glyph_type)
        print()


def ls(session: Session, *, is_long: bool = False, show_context: bool = False):
    all_nodes = _get_all_nodes(session)
    for (view_name, ctx), (index, host, view, node) in all_nodes.items():
        print(f"{index}  ", end="")
        if is_long:
            path = _get_node_path(host, view, view_name, node)
            print(f"{path}", end="")
        else:
            print(f"{node.props.get('Name')}", end="")

        if show_context:
            print(f"    {node.ctx}")
        else:
            print()


def _print_info(*, host: str, view_name: str, view: ViewInfo, node: Node):
    if not node.props.get("Name"):
        return

    print("ctx:", node.ctx)
    print("host:", host)
    print("view:", view_name)
    path = _get_node_path(host, view, view_name, node)
    print("path:", path)

    props_dict = node.props
    for key, val in props_dict.items():
        if key != "regs":
            print(f"{key}: {val}")
    # Special printing for all regs - they need to be refreshed
    regs = node.props.get("regs")
    if regs:
        reg_names_to_refresh = tuple(regs.keys())
        node.update_regs(reg_names=reg_names_to_refresh, force=True, done=None)
        for reg_key, reg_val in regs.items():
            reg_data_len = len(reg_val.data)
            for index in range(0, reg_data_len):
                int_value = int.from_bytes(reg_val.data[index], byteorder="little", signed=False)
                if reg_data_len == 1:
                    print(f"regs.{reg_key:}: {hex(int_value)}")
                else:
                    print(f"regs.{reg_key:}.slr{index}: {hex(int_value)}")


def info(session: Session, *, match_string: str):
    all_nodes = _get_all_nodes(session)
    try:
        match_string_int_value = int(match_string)
    except ValueError:
        match_string_int_value = -1

    for (view_name, ctx), (index, host, view, node) in all_nodes.items():
        if node.ctx == match_string or index == match_string_int_value:
            _print_info(host=host, view_name=view_name, view=view, node=node)
            break


def parse_args(cmdline_args):
    global _hw_url, _cs_url

    parser = argparse.ArgumentParser(description="ChipScoPy CLI")
    parser.add_argument("-v", "--version", action="store_true", help="Display version")
    parser.add_argument("--no-banner", action="store_true", help="Turn off banner text")
    parser.add_argument("--hw_url", type=str, help="Hardware Server URL")
    parser.add_argument("--cs_url", type=str, help="ChipScope Server URL")

    base_parser = argparse.ArgumentParser(add_help=False)

    subparsers = parser.add_subparsers(help="sub-command", dest="subcommand")

    report_parser = subparsers.add_parser("report", help="report help", parents=[base_parser])
    report_parser.add_argument("--devices", action="store_true", help="Report devices")
    report_parser.add_argument("--servers", action="store_true", help="Report server versions")

    info_parser = subparsers.add_parser("info", help="info help", parents=[base_parser])
    info_parser.add_argument("context", type=str)

    ls_parser = subparsers.add_parser("ls", help="ls help", parents=[base_parser])
    ls_parser.add_argument(
        "--show-context", required=False, action="store_true", help="Print additional node context"
    )
    ls_parser.add_argument(
        "--long",
        "-l",
        required=False,
        action="store_true",
        help="long listing including device contexts",
    )

    program_parser = subparsers.add_parser("program", help="program help", parents=[base_parser])
    program_parser.add_argument("file", type=str, nargs="?")
    program_parser.add_argument("--dna", type=str, help="find device to program with dna")
    program_parser.add_argument("--cable-context", type=str, help="jtag cable context")
    program_parser.add_argument("--context", type=str, help="device context")
    program_parser.add_argument("--part", type=str, help="part name")
    program_parser.add_argument("--family", type=str, help="family name")
    program_parser.add_argument(
        "--jtag-index", type=int, help="jtag index in chain (starting index is 0)"
    )
    program_parser.add_argument(
        "--device-index", type=int, help="device index device list (starting index is 0)"
    )
    program_parser.add_argument(
        "--skip-reset", action="store_true", help="Skip the reset before programming device"
    )
    program_parser.add_argument("--list", action="store_true", help="list devices", dest="list_")
    program_parser.add_argument(
        "--program-log", action="store_true", help="download programming log"
    )
    program_parser.add_argument("--force-reset", action="store_true", help="reset device")

    tree_parser = subparsers.add_parser("tree", help="tree help", parents=[base_parser])
    tree_parser.add_argument(
        "--view",
        type=str,
        default=None,
        required=False,
        choices=["jtag", "memory", "debugcore", "chipscope"],
        help="Use the specified view for walking nodes",
    )
    tree_parser.add_argument(
        "--show-context", required=False, action="store_true", help="Print additional node context"
    )
    tree_parser.add_argument(
        "--glyph-type",
        type=str,
        default="std",
        required=False,
        choices=["std", "alt", "space", "none"],
        help="Tree glyph style",
    )
    tree_parser.add_argument(
        "--properties",
        action="store_true",
        help="Show properties for nodes",
    )

    args = parser.parse_args(cmdline_args)

    if args.version:
        print(chipscopy.__version__)
        sys.exit(0)

    if not args.no_banner:
        display_banner()

    if args.hw_url:
        _hw_url = args.hw_url

    if args.cs_url:
        _cs_url = args.cs_url

    if args.subcommand == "report":
        with create_session(
            hw_server_url=_hw_url, cs_server_url=_cs_url, bypass_version_check=_bypass_version_check
        ) as session:
            report(session=session, devices=args.devices, servers=args.servers)

    elif args.subcommand == "info":
        with create_session(
            hw_server_url=_hw_url,
            cs_server_url=_cs_url,
            initial_device_scan=False,
            bypass_version_check=_bypass_version_check,
        ) as session:
            info(session=session, match_string=args.context)

    elif args.subcommand == "ls":
        with create_session(
            hw_server_url=_hw_url,
            cs_server_url=_cs_url,
            initial_device_scan=False,
            bypass_version_check=_bypass_version_check,
        ) as session:
            ls(session=session, is_long=args.long, show_context=args.show_context)

    elif args.subcommand == "program":
        with create_session(
            hw_server_url=_hw_url, cs_server_url=_cs_url, bypass_version_check=_bypass_version_check
        ) as session:
            program(
                session=session,
                file=args.file,
                dna=args.dna,
                cable_context=args.cable_context,
                context=args.context,
                part=args.part,
                family=args.family,
                jtag_index=args.jtag_index,
                device_index=args.device_index,
                skip_reset=args.skip_reset,
                list_=args.list_,
                program_log=args.program_log,
                force_reset=args.force_reset,
            )

    elif args.subcommand == "tree":
        with create_session(
            hw_server_url=_hw_url,
            cs_server_url=_cs_url,
            initial_device_scan=False,
            bypass_version_check=_bypass_version_check,
        ) as session:
            tree(
                session=session,
                view=args.view,
                show_context=args.show_context,
                glyph_type=args.glyph_type,
                print_properties=args.properties,
            )


def main():  # pragma: no cover
    parse_args(sys.argv[1:])


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, ClassVar, Dict
from dataclasses import dataclass

from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.tcf.services import DoneHWCommand


@dataclass
class Sysmon(DebugCore["SysMonCoreClient"]):
    """Sysmon is the top level API for accessing System Monitor in Versal."""

    def __init__(self, sysmon_tcf_node):
        super(Sysmon, self).__init__(CoreType.SYSMON, sysmon_tcf_node)

        self.name = sysmon_tcf_node.props["Name"]

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name}

        self.schedule = None

    def __str__(self):
        return self.name

    def initialize_sensors(self, done: DoneHWCommand = None):
        """
        Initializes all of the server side constructs for sysmon, required for most subsequent APIs.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            No data is returned directly from this method.

        """
        self.core_tcf_node.initialize_sensors(done)

    def refresh_measurement_schedule(self, done: DoneHWCommand = None) -> Dict[int, str]:
        """
        Dynamically scans the SysMon to determine the sensors scheduled into the system, in order.


        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a dict where the keys are the root IDs and the values are the named sensors.

        """
        schedule = self.core_tcf_node.refresh_measurement_schedule(done)
        self.schedule = schedule
        return schedule

    def get_measurements(self, done: DoneHWCommand = None) -> List[object]:
        """
        Advanced API

        Returns the measurement objects

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a list of measurement configuration objects

        """
        sensors = self.core_tcf_node.get_measurements(done)
        return sensors

    def get_all_sensors(self, done: DoneHWCommand = None) -> dict:
        """
        This function is identical to ``get_supported_sensors`` in customer releases. For VnC teams this API offers
        internal sensors. For information on how to use this API please contact @dkopelov or #chipscope.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            A list of strings representing all sensors for the connected device.

        """
        sensors = self.core_tcf_node.get_all_sensors(done)
        return sensors

    def configure_measurement_schedule(
        self, measurements: Dict[str, str], done: DoneHWCommand = None
    ) -> None:
        self.core_tcf_node.configure_measurement_schedule(measurements, done)

    def stream_sensor_data(self, interval: int, done: DoneHWCommand = None) -> None:
        """
        This API configures polling in hw_server to stream the data from the measurement schedule (including device
        temps) at the specified ``interval``.

        To process the incoming data, a `dm.NodeListener` must be declared and registered on this node.


        node_listener = dm.NodeListener(sysmon)

        session.chipscope_view.add_node_listener(node_listener)

        Args:
            interval: poll period (in ms)
            done: Optional command callback that will be invoked when the response is received.
        """
        self.core_tcf_node.stream_sensor_data(interval, done)

    def read_sensor(self, sensor: str) -> float:
        """
        Read single sensor value from hw
        Args:
            sensor:  the name of the sensor to read

        Returns:
            refreshed value of specific sensor (float) or None if sensor isn't detected in design

        """

        if self.schedule is None:
            print(
                f"warning: read_sensor(): measurement schedule is empty or has not been refreshed"
            )
            return None

        if sensor not in self.schedule.values():
            print(
                f"warning: read_sensor(): requested sensor {sensor} is not in the measurement schedule"
            )
            return None

        supply_idx = None
        for idx, s in self.schedule.items():
            if sensor == s:
                supply_idx = idx
                break

        supply_name = f"supply{supply_idx}"
        # greedy
        current_value = self.core_tcf_node.refresh_property(supply_name)
        return current_value[supply_name]

    def read_sensors(self, sensors: List[str]) -> Dict[str, float]:
        """
        Read a list of sensor values from hw
        Args:
            sensors:  list of the names of the sensors to read

        Returns:
            refreshed values of specific sensors (float)
            if the sensor isn't sequenced in the design then a warning msg is printed and the sensor is dropped from
            the list
        """
        props_to_refresh = []
        out_data = {}
        supply_x_map = {}
        if self.schedule is None:
            print(
                f"warning: read_sensors(): measurement schedule is empty or has not been refreshed"
            )
            return None

        for sensor in sensors:
            if sensor not in self.schedule.values():
                print(
                    f"warning: read_sensors(): requested sensor {sensor} is not in the measurement schedule"
                )
                continue

            supply_idx = None
            for idx, s in self.schedule.items():
                if sensor == s:
                    supply_idx = idx
                    break

            out_data[sensor] = None
            supply_name = f"supply{supply_idx}"
            supply_x_map[supply_name] = sensor
            props_to_refresh.append(supply_name)

        current_sensor_values = self.core_tcf_node.refresh_property(props_to_refresh)
        for supply_name, value in current_sensor_values.items():
            out_data[supply_x_map[supply_name]] = value

        return out_data

    def read_temp(self) -> Dict[str, float]:
        """
        Read temp values from hw
        Args:
            None

        Returns:
            refreshed value of device_temp, max, and min in Celsius
        """
        temp_dict = self.core_tcf_node.refresh_property(
            ["device_temp", "device_temp_max_max", "device_temp_min_min"]
        )
        return temp_dict



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import sys
from dataclasses import dataclass, field
from enum import IntEnum
from functools import reduce
from io import TextIOBase
from typing import List, Dict, Optional, Union

from antlr4 import ParserRuleContext
from antlr4.error.ErrorListener import ErrorListener
from antlr4.tree.Tree import TerminalNodeImpl

from chipscopy.api.ila import ILAPort, ILAProbe


class ILATsmErrorListener(ErrorListener):
    def __init__(self, out_stream: TextIOBase = sys.stderr):
        self._out_stream: TextIOBase = out_stream
        self._error_count = 0

    def get_error_count(self) -> int:
        return self._error_count

    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        self.report_error(line, column, msg)

    def report_error(self, line, column, msg):
        self._error_count += 1
        if line is None or column is None:
            print(msg, file=self._out_stream)
        else:
            print("[Line " + str(line) + ":" + str(column) + "] " + msg, file=self._out_stream)

    def report_error_ctx(self, ctx: Union[ParserRuleContext, TerminalNodeImpl], msg: str):
        if isinstance(ctx, ParserRuleContext):
            self.syntaxError(None, None, ctx.start.line, ctx.start.column, msg, None)
        else:
            self.syntaxError(None, None, ctx.symbol.line, ctx.symbol.column, msg, None)


# TSM constants
COUNTER_COUNT = 4
TABLE_ROWS_PER_STATE = 8
MAX_STATE_COUNT = 16
MAX_BRANCH_COUNT_PER_STATE = 3
MAX_CONDITION_COUNT_PER_STATE = MAX_BRANCH_COUNT_PER_STATE - 1


@dataclass
class ILATsmActionPair:
    name: str
    index: int

    def __str__(self):
        return self.name + str(self.index)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.name == other.name and self.index == other.index


class ILATsmInput(IntEnum):
    CUR_STATE_3 = 6
    CUR_STATE_2 = 5
    CUR_STATE_1 = 4
    CUR_STATE_0 = 3

    IF_COND_MUX = 2
    ELSE_IF_MUX = 1

    COUNTER_CMP_MUX = 0

    NUM_INPUTS = 7


class ILATsmAction(enum.Flag):
    # FSM ILA core register action encoding.
    DEFAULT = 0x00_0000

    GOTO_0 = 0x00_0001
    GOTO_1 = 0x00_0002
    GOTO_2 = 0x00_0004
    GOTO_3 = 0x00_0008

    RESET_COUNTER_0 = 0x00_0010
    INCREMENT_COUNTER_0 = 0x00_0020

    RESET_COUNTER_1 = 0x00_0040
    INCREMENT_COUNTER_1 = 0x00_0080

    RESET_COUNTER_2 = 0x00_0100
    INCREMENT_COUNTER_2 = 0x00_0200

    RESET_COUNTER_3 = 0x00_0400
    INCREMENT_COUNTER_3 = 0x00_0800

    SELECT_COUNTER_1_OR_3 = 0x00_1000
    SELECT_COUNTER_2_OR_3 = 0x00_2000

    SET_FLAG_0 = 0x00_4000
    SET_FLAG_1 = 0x00_8000
    SET_FLAG_2 = 0x01_0000
    SET_FLAG_3 = 0x02_0000

    CAPTURE = 0x04_0000  # Not supported
    TRIGGER = 0x08_0000

    CLEAR_FLAG_0 = 0x10_0000
    CLEAR_FLAG_1 = 0x20_0000
    CLEAR_FLAG_2 = 0x40_0000
    CLEAR_FLAG_3 = 0x80_0000

    @staticmethod
    def create(action: ILATsmActionPair):
        if action.name == "goto":
            val = action.index
        elif action.name == "reset_counter":
            val = ILATsmAction.RESET_COUNTER_0.value << (action.index * 2)
        elif action.name == "increment_counter":
            val = ILATsmAction.INCREMENT_COUNTER_0.value << (action.index * 2)
        elif action.name == "set_flag":
            val = ILATsmAction.SET_FLAG_0.value << action.index
        elif action.name == "clear_flag":
            val = ILATsmAction.CLEAR_FLAG_0.value << action.index
        elif action.name == "trigger":
            val = ILATsmAction.TRIGGER.value
        else:
            val = 0
        return ILATsmAction(val)

    @staticmethod
    def from_action_pair_list(actions: List[ILATsmActionPair]):
        fsm_actions = map(ILATsmAction.create, actions)
        combined_actions = reduce((lambda x, y: x | y), fsm_actions)
        return combined_actions

    @staticmethod
    def add_select_counter_action(action, counter_index: int):
        """
        Adds to input "action" value the select_counter actions corresponding to "counter_index".
        Args:
            action (ILATsmAction): Old action value.
            counter_index (int): Counter index corresponding to the next state.

        Returns: New action object which has input "action" actions plus SELECT_COUNTER actions.

        """
        val = action.value
        if counter_index == 1 or counter_index == 3:
            val |= ILATsmAction.SELECT_COUNTER_1_OR_3.value
        if counter_index == 2 or counter_index == 3:
            val |= ILATsmAction.SELECT_COUNTER_2_OR_3.value
        return ILATsmAction(val)

    def dump(self) -> str:
        ITA = ILATsmAction
        goto_state = self.value & 0xF
        select_counter = 1 if ITA.SELECT_COUNTER_1_OR_3 in self else 0
        select_counter += 2 if ITA.SELECT_COUNTER_2_OR_3 in self else 0
        mask = (
            ITA.SELECT_COUNTER_1_OR_3
            | ITA.SELECT_COUNTER_2_OR_3
            | ITA.GOTO_0
            | ITA.GOTO_1
            | ITA.GOTO_2
            | ITA.GOTO_3
        )
        aa = ILATsmAction(self.value & ~mask.value)
        s = str(aa) + f"|SEL_C:{select_counter}|GOTO:{goto_state}"
        # Clean up string.
        s = s.replace("FLAG_", "F:")
        s = s.replace("COUNTER_", "C:")
        s = s.replace("ILATsmAction.", "")
        s = s.replace("DEFAULT", "")
        s = s.replace("INCREMENT", "INC")
        s = s.replace("|", " ")

        bin_value = f"{self.value:024b}"
        bin_value = "_".join(bin_value[idx : idx + 4] for idx in range(0, len(bin_value), 4))

        return bin_value + " " + s.lower()


@dataclass
class ILATsmCounterCompare:
    counter_index: int
    compare_op: str
    compare_value: str

    def __str__(self):
        return f"$COUNTER{self.counter_index} {self.compare_op} {self.compare_value}"

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return (
            self.counter_index == other.counter_index
            and self.compare_op == other.compare_op
            and self.compare_value == other.compare_value
        )


@dataclass
class ILATsmMatchUnit:
    # Reverse order example: "16'H0004 <= probe_a", instead of "probe_a > 16'H0004"
    probe: ILAProbe
    compare_op: str
    compare_value: str
    reverse_order: bool

    def get_compare_operator(self) -> str:
        return reverse_compare_operator(self.compare_op) if self.reverse_order else self.compare_op


@dataclass
class ILATsmCondition:
    probe_operator: Optional[str]
    counter_operator: Optional[str]
    counter_compare: Optional[ILATsmCounterCompare]
    match_units: List[ILATsmMatchUnit] = field(default_factory=lambda: [])

    def is_invert_counter_mux(self) -> bool:
        return self.counter_compare and self.counter_compare.compare_op == "!="

    def get_mapper_probe_operator(self) -> str:
        # "and" is default value.
        return "or" if self.probe_operator == "||" else "and"

    @staticmethod
    def from_item_list(condition_list: List[Union[str, ILATsmMatchUnit, ILATsmCounterCompare]]):
        """condition_list (): List made up of "||", "&&", ILATsmMatchUnit, ILATsmCounterCompare."""
        probe_op, counter_op, counter_compare = None, None, None
        if len(condition_list) == 1:
            # Handle single ILATsmCounterCompare item.
            # Single ILATSmMatchUnit item case is handled by collecting "mus" below.
            if isinstance(condition_list[0], ILATsmCounterCompare):
                counter_compare = condition_list[0]
        else:
            for op_idx in range(1, len(condition_list), 2):
                pre = condition_list[op_idx - 1]
                op = condition_list[op_idx]
                post = condition_list[op_idx + 1]
                if isinstance(pre, ILATsmCounterCompare):
                    counter_op = op
                    counter_compare = pre
                elif isinstance(post, ILATsmCounterCompare):
                    counter_op = op
                    counter_compare = post
                elif isinstance(pre, ILATsmMatchUnit) and isinstance(post, ILATsmMatchUnit):
                    probe_op = op

        mus = list(filter(lambda x: isinstance(x, ILATsmMatchUnit), condition_list))
        return ILATsmCondition(probe_op, counter_op, counter_compare, mus)


@dataclass
class ILATsmBranch:
    index: int
    actions: ILATsmAction
    condition: Optional[ILATsmCondition]
    next_state: int


@dataclass
class ILATsmState:
    name: str
    index: int
    counter_index: int = 0
    branches: List[ILATsmBranch] = field(default_factory=lambda: [])


@dataclass
class ILATsmData:
    name: str
    ports: List[ILAPort]
    probes: Dict[str, ILAProbe]
    counter_widths: List[int]
    states: List[ILATsmState] = field(default_factory=lambda: [])

    def get_match_unit_count(self):
        return sum(len(port.match_units) for port in self.ports)


def reverse_compare_operator(op: str) -> str:
    """
    Return operator when reversing order of operands.
    E.g.  16'h4000 < $counter3 => $counter3 >= 16'h4000
    """
    reverse = {">": "<=", "<": ">=", ">=": "<", "<=": ">"}
    # For "==" and "!=", the reverse is the same as the passed in operator.
    return reverse.get(op, op)



# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from typing import NewType
import threading
import inspect
import importlib
from .. import protocol
from ...channel import *
import pytcf

_providers = []
_lock = threading.RLock()

TokenType = NewType("TokenType", Token)


class TCFError(Exception):
    pass


class DoneHWCommand(object):
    """Client call back interface for generic commands."""
    def doneHW(self, token, error, args):
        """Called when memory operation command command is done.
        :param token: Command handle.
        :param error: Error object or **None**.
        """
        pass

    def __call__(self, token, error, args):
        return self.doneHW(token, error, args)


class ServiceProvider(object):
    """Clients can implement this abstract class if they want to provide
    implementation of a local service or remote service proxy.
    """

    def get_local_service(self, channel):
        pass

    def get_service_proxy(self, channel, service_name):
        pass


def add_service_provider(provider):
    with _lock:
        _providers.append(provider)


def removeServiceProvider(provider):
    with _lock:
        _providers.remove(provider)


def onChannelCreated(channel, services_by_name):
    with _lock:
        for provider in _providers:
            try:
                arr = provider.get_local_service(channel)
                if not arr:
                    continue
                for service in arr:
                    if service.getName() in services_by_name:
                        continue
                    services_by_name[service.getName()] = service
            except Exception as x:
                protocol.log("Error calling TCF service provider", x)


def onChannelOpened(channel, service_names, services_by_name):
    with _lock:
        for name in service_names:
            for provider in _providers:
                try:
                    service = provider.get_service_proxy(channel, name)
                    if not service:
                        continue
                    services_by_name[name] = service
                    break
                except Exception as x:
                    protocol.log("Error calling TCF service provider", x)
            if name in services_by_name:
                continue
            services_by_name[name] = GenericProxy(channel, name)


def getServiceManagerID():
    # In current implementation ServiceManager is a singleton,
    # so its ID is same as agent ID.
    return protocol.getAgentID()


class GenericCallback(object):

    def __init__(self, callback):
        self.callback = callback

    def __getattr__(self, attr):
        if attr.startswith("done"):
            return self.callback


class Service(object):
    """TCF service base class."""
    def getName(self):
        """Abstract method to get the service name.

        :returns: This service name
        """
        raise NotImplementedError("Abstract method")

    def __init__(self, channel):
        self.channel = channel

    def __str__(self):
        """TCF service string representation.

        :returns: The name of the service.
        """
        return self.getName()

    def _makeCallback(self, done):
        """Turn *done* into a callable.

        If *done* is already a :class:`collections.Callable`, it is returned
        as is, else, it is made callable, and returned.

        :param done: The item to make callable.

        :returns: The callable value of *done*
        """
        if inspect.isfunction(done) or inspect.ismethod(done):
            return GenericCallback(done)
        return done

    def send_command(self, cmd, args, done=None, progress=None):
        if not pytcf.is_dispatch_thread():
            return pytcf.post_event_and_wait(self.send_command, cmd, args, done, progress)
        token = None

        def reply_handler(chan, error):
            if not error:
                data = pytcf.read_command_args(chan)
                error = data[0] if len(data) > 0 else error
                data = data[1:] if len(data) > 1 else ()
            else:
                data = ()

            if done:
                done(token, error, data)

        def progress_handler(chan):
            data = pytcf.read_command_args(chan)
            if progress:
                progress(token, data)

        token = pytcf.send_command(
            self.channel,
            self.getName(),
            cmd,
            reply_handler if done else None,
            progress_handler if progress else None)
        pytcf.write_command_args(self.channel, args)
        return token

    def send_xicom_command(self, cmd, args, done, progress=None):
        if not pytcf.is_dispatch_thread():
            return pytcf.post_event_and_wait(self.send_command, cmd, args, done, progress)
        token = None

        def reply_handler(chan, error):
            if not error:
                data = pytcf.read_command_args(chan)
                error = data[0] if len(data) > 0 else error
                if isinstance(error, dict) and 'Format' in error:
                    e = Exception(error['Format'])
                    try:
                        mod = importlib.import_module(error['Module'])
                        cls = getattr(mod, error["Class"])
                        e = cls(f"TCF Error: {self.getName()} {cmd}: {error['Format']}")
                    except Exception:
                        pass
                    error = e
                data = data[1:] if len(data) > 1 else ()
                if data and len(data) == 1:
                    data = data[0]
            else:
                data = ()

            if done:
                done(token, error, data)

        def progress_handler(chan):
            data = pytcf.read_command_args(chan)
            if progress:
                progress(token, data)

        token = pytcf.send_command(
            self.channel,
            self.getName(),
            cmd,
            reply_handler if done else None,
            progress_handler if progress else None)
        pytcf.write_command_args(self.channel, args)
        return token

    def send_old_xicom_command(self, cmd, args, done):
        if not pytcf.is_dispatch_thread():
            return pytcf.post_event_and_wait(self.send_command, cmd, args, done)
        token = None

        def reply_handler(chan, error):
            if not error:
                data = pytcf.read_command_args(chan)
                # error = data[0] if len(data) > 0 else error
                # data = data[1:] if len(data) > 1 else ()
            else:
                data = ()

            if done:
                done(token, error, data)

        token = pytcf.send_command(
            self.channel,
            self.getName(),
            cmd,
            reply_handler if done else None,
            None)
        pytcf.write_command_args(self.channel, args, True)
        return token

    send_old_command = send_old_xicom_command


class ServiceSync(object):
    def __init__(self, channel, service_name, raise_on_error=False):
        self.service_name = service_name
        self.chan = channel
        self.result = None
        self.raise_on_error = raise_on_error

        #services = protocol.invokeAndWait(self.chan.getRemoteServices)
        # if service_name not in services:
        #     raise Exception('No service {} found'.format(service_name))

    def __getattr__(self, attr):
        return CommandWrapper(self, attr)

    def _doneCommand(self, token, error, args):
        self.result = CommandResult(token, error, args)

    def invoke(self, command, *args, **kwargs):
        cond = threading.Condition()
        self.result = None

        def callCommand(command, *args, **kwargs):
            serv = self.chan.getRemoteService(self.service_name)

            def doneHW(token, error, args):
                self._doneCommand(token, error, args)
                with cond:
                    cond.notify()

            try:
                kwargs['done'] = doneHW
                f = getattr(serv, command, None)
                f(*args, **kwargs)
            except TypeError as e:
                error = '{} service {} error: {}'.format(serv, command, str(e))
                self._doneCommand(Token(), error, [])
                with cond:
                    cond.notify()

        if protocol.isDispatchThread():
            callCommand(command, *args, **kwargs)
        else:
            with cond:
                protocol.invokeLater(callCommand, command, *args, **kwargs)
                # When inside the debugger wait longer, don't time out.
                # todo: move this feature to the command line???
                cond.wait(None if __debug__ else 60)

        if self.raise_on_error and self.result and self.result.error:
            if isinstance(self.result.error, Exception):
                raise self.result.error
            else:
                raise Exception(self.service_name + ':' + command + ': ' + str(self.result.error))
        return self.result


class CommandResult(object):
    def __init__(self, token, error, args):
        self.token = token
        self.error = error
        # unwrap result if only one element
        # if args and len(args) == 1 and \
        #         not isinstance(args, dict):
        #     args = args[0]
        self.args = args

    def __str__(self):
        if self.error:
            return "[%s] error: %s" % (self.token.id, self.error)
        return "[%s] result: %s" % (self.token.id, self.args)
    __repr__ = __str__

    def __iter__(self):
        yield self.error
        yield self.args

    def get(self, key=None):
        if key:
            return self.args.get(key)
        return self.args


class CommandWrapper(object):
    def __init__(self, control, command):
        self._control = control
        self._command = command

    def __call__(self, *args, **kwargs):
        return self._control.invoke(self._command, *args, **kwargs)


class GenericProxy(Service):
    """Objects of GenericProxy class represent remote services, which don't
    have a proxy class defined for them.

    Clients still can use such services, but framework will not provide
    service specific utility methods for message formatting and parsing.
    """

    def __init__(self, channel, name):
        self.__channel = channel
        self.name = name

    def getName(self):
        return self.name

    def getChannel(self):
        return self.__channel
class DefaultServiceProvider(ServiceProvider):
    package_base = str(__package__) + ".remote"

    def get_local_service(self, channel):
        # TODO DiagnosticsService
        # return [DiagnosticsService(channel)]
        return []

    def get_service_proxy(self, channel, service_name):
        # package_base = str(__package__) + "...services.remote"
        package_base = "chipscopy.tcf.services.remote"
        service = None
        try:
            # Xicom_v1.XX special case
            if 'Xicom_v1' in service_name:
                service_name = "Xicom"
            clsName = service_name + "Proxy"
            clsName = clsName.replace('.', '_')
            package = package_base + "." + clsName
            clsModule = __import__(package, fromlist=[clsName],
                                   globals=globals())
            cls = clsModule.__dict__.get(clsName)
            service = cls(channel)
            # assert service_name == service.getName()
        except ImportError:
            pass
        except Exception as x:
            protocol.log("Cannot instantiate service proxy for " +
                         service_name, x)
        return service


add_service_provider(DefaultServiceProvider())



# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
TCF - Target Communication Framework
"""

import types

from .util import task

__all__ = ('connect', 'peers')

try:
    import pytcf
    import sys
    from .native import protocol, services
    from .services import locator
    from pprint import pprint
    sys.modules["chipscopy.tcf.protocol"] = protocol
    sys.modules["chipscopy.tcf.services"] = services
    sys.modules["chipscopy.tcf.services.locator"] = locator

    def connect(params, wait=True):
        """Connect to peer. Argument is a string of the form
        <transport>:<host>:<port>, e.g. "TCP:127.0.0.1:1534".
        """
        if isinstance(params, compat.strings):
            params = _parse_params(params)
        elif isinstance(params, dict):
            raise TypeError("Expected string or dict")
        return pytcf.connect_tcf(params[peer.ATTR_ID])

except ModuleNotFoundError:
    from . import compat, protocol, peer, channel

    def connect(params, wait=True):
        """Connect to peer. Argument is a string of the form
        <transport>:<host>:<port>, e.g. "TCP:127.0.0.1:1534".
        """
        if isinstance(params, compat.strings):
            params = _parse_params(params)
        elif isinstance(params, dict):
            raise TypeError("Expected string or dict")
        p = peer.TransientPeer(params)
        if wait:
            c = task.Task(_openChannel, p).get()
        else:
            c = protocol.invokeAndWait(p.openChannel)
        return c


    def peers():
        "Return list of discovered remote peers"
        locator = protocol.getLocator()
        if locator:
            return protocol.invokeAndWait(locator.getPeers)


    def _openChannel(p, done=None):
        assert protocol.isDispatchThread()
        c = p.openChannel()
        if done is None:
            return

        class ChannelListener(channel.ChannelListener):
            def onChannelOpened(self):
                c.removeChannelListener(self)
                done(None, c)

            def onChannelClosed(self, error):
                done(error, None)
        c.addChannelListener(ChannelListener())


def process_param_str(params, default_port="3121"):
    parts = params.split(":", 3)
    if len(parts) < 3:
        parts.insert(0, "TCP")
    if len(parts) < 3:
        parts.append(default_port)
    return ":".join(parts)


def _parse_params(paramStr):
    args = process_param_str(paramStr).split(":")
    if len(args) != 3:
        raise ValueError("Expected format: <transport>:<host>:<port>")
    transp, host, port = args
    return {
        peer.ATTR_ID: paramStr,
        peer.ATTR_IP_HOST: host,
        peer.ATTR_IP_PORT: port,
        peer.ATTR_TRANSPORT_NAME: transp
    }



# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *     Xilinx
# *****************************************************************************

try:
    import pytcf
    from ...native.services.local import *
except:
    import re
    import time
    from typing import ClassVar
    from ... import channel, transport
    from ...channel import fromJSONSequence, toJSONSequence
    from .. import to_xargs, ServiceProvider, Service, add_service_provider


    class CommandRequest(object):
        def __init__(self, _channel, token, data, node=None):
            self.channel = _channel
            self.token = token
            self.data = data
            if node is not None:
                self.node = node
            self.args = fromJSONSequence(data) if data else None

        def sendError(self, error, *result):
            if type(error) is list:
                error_props = {
                    "Code": 0x20001,  # ERR_OTHER default used by hw_server
                    "Time": int(time.perf_counter() * 1000),
                    "Format": str(error),
                }
            else:
                error_props = {
                    "Code": 0x20001,  # ERR_OTHER default used by hw_server
                    "Time": int(time.perf_counter() * 1000),
                    "Format": str(error),
                    "Class": error.__class__.__name__,
                    "Module": error.__class__.__module__
                }
            self.channel.sendResult(self.token, toJSONSequence((error_props, *result)))

        def sendResult(self, *result):
            self.channel.sendResult(self.token, toJSONSequence((None, *result)))

        def sendXicomResult(self, *results):
            result = to_xargs(results)
            self.channel.sendResult(self.token, toJSONSequence((None, *result)))

        def sendXicomProgress(self, *progress):
            args = to_xargs(progress)
            self.channel.sendProgress(self.token, toJSONSequence((None, *args)))


    def send_xicom_event(service_name, event_name, args):
        data = toJSONSequence(to_xargs(args))
        transport.sendEvent(service_name, event_name, data)


    def to_snake_case(name: str) -> str:
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


    class CommandServer(channel.CommandServer):
        def __init__(self, service, _channel):
            self.service = service
            self.channel = _channel

        def command(self, token, name, data):
            cmd = getattr(self.service, to_snake_case(name) + "_cmd", None)
            if not cmd:
                cmd = getattr(self.service, name + 'Cmd', None)
            if cmd:
                try:
                    request = CommandRequest(self.channel, token, data)
                    cmd(request)
                except Exception as x:
                    msg = f'ChipScope Service {self.service.getName()} {name}: {x}.'
                    request.sendError(msg)
            else:
                channel.rejectCommand(token)


    class LocalServiceProvider(ServiceProvider):
        def __init__(self, service):
            self.service = service

        def get_local_service(self, _channel):
            _channel.addCommandServer(
                self.service,
                CommandServer(self.service, _channel)
            )
            return self.service,


    _local_services_added = set()


    def add_local_service(service: Service or ClassVar[Service]):
        if str(service) not in _local_services_added:
            add_service_provider(LocalServiceProvider(service))
            _local_services_added.add(str(service))



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from abc import ABC, abstractmethod
from typing import Dict, Optional

from chipscopy.client.view_info import ViewInfo
from chipscopy.dm import Node
from chipscopy.client.jtagdevice import JtagCable
from chipscopy.client import ServerInfo
from chipscopy.api.device.device import Device, discover_devices
from chipscopy.api.containers import QueryList

DOMAIN_NAME = "client"


class Cable(ABC):
    """A Cable represents a cable connection. This may be a physical jtag cable, or a virtual cable.
    Zero or more devices may be associated with the cable, and each device has a position.
    Roughly equivalent to a vivado hw_target.
    """

    def __init__(self):
        pass

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return self.to_json()

    def __getitem__(self, item):
        props = self.to_dict()
        if item in props:
            return props[item]
        else:
            raise AttributeError(f"No property {str(item)}")

    @property
    @abstractmethod
    def is_valid(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def devices(self) -> QueryList[Device]:
        raise NotImplementedError

    @abstractmethod
    def to_json(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> Dict:
        raise NotImplementedError

    @property
    @abstractmethod
    def context(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


class _VirtualCable(Cable):
    # TODO: Placeholder - Implement when needed
    pass


class _JtagCable(Cable):
    def __init__(
        self,
        tcf_node: Node,
        hw_server: ServerInfo,
        cs_server: ServerInfo,
        disable_core_scan: bool,
        timeout: int,
    ):
        super().__init__()
        self._tcf_node = tcf_node
        self._hw_server = hw_server
        self._cs_server = cs_server
        self._disable_core_scan = disable_core_scan
        self._timeout = timeout
        self._should_wait_for_ready = True
        assert tcf_node.props.get("node_cls") == JtagCable
        assert tcf_node.is_valid is True

    @property
    def is_valid(self) -> bool:
        return self._tcf_node.is_valid

    @property
    def devices(self) -> QueryList[Device]:
        if self._should_wait_for_ready:
            # Only wait first time
            self._should_wait_for_ready = False
            self.wait_for_cable_ready(self._timeout)
        devices = discover_devices(
            hw_server=self._hw_server,
            cs_server=self._cs_server,
            disable_core_scan=self._disable_core_scan,
            cable_ctx=self._tcf_node.ctx,
        )
        return devices

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=4, default=lambda o: str(o))

    def to_dict(self):
        d = {}
        d.update(self._tcf_node.props)
        d.update({"context": self._tcf_node.ctx})
        return d

    @property
    def context(self) -> str:
        return self._tcf_node.ctx

    @property
    def name(self) -> str:
        return self._tcf_node.props.get("Description")

    def wait_for_cable_ready(self, cable_timeout: int):
        """Waits for a cable to be ready up to cable_timeout seconds.

        Args:
            cable_timeout: Max seconds to wait
        """
        if cable_timeout > 0:
            view = self._hw_server.get_view("jtag")
            for ctx, cable in _get_cable_contexts(view).items():
                if ctx == self.context:
                    # Wait for cable to become ready
                    start = time.time()
                    while time.time() - start < cable_timeout:
                        device_nodes = list(view.get_children(cable))
                        if len(device_nodes) > 0:
                            break
                        time.sleep(0.25)


def _get_cable_contexts(view: ViewInfo) -> Dict[str, Node]:
    """Returns a dict of all cable contexts mapped to cable nodes attached to the hw_server"""
    cables = view.get_children()
    cable_ctx_dict = {}
    for cable in cables:
        cable_ctx_dict[cable.ctx] = cable
    return cable_ctx_dict


def wait_for_all_cables_ready(hw_server: ServerInfo, cable_timeout: int):
    """Wait for all jtag cables to be ready. Simple method - iterate over all the cables. If any
    have 0 devices, wait until devices appear. If we get a timeout, there are assumed to be
    no devices on the cable.

    Some cables take a second or two to come up. Don't have a good way to ask the
    cable if it is ready, so we just inspect the device nodes and see if they exist.

    Args:
        cable_timeout: max timeout in seconds to initialize a cable
        hw_server: hardware server
    """
    if cable_timeout > 0:
        view = hw_server.get_view("jtag")
        start = time.time()
        keep_trying = True
        cables = _get_cable_contexts(view)
        while keep_trying and time.time() - start < cable_timeout:
            keep_trying = False
            for ctx, cable in cables.items():
                # Wait for cable to become ready
                device_nodes = list(view.get_children(cable))
                if len(device_nodes) == 0:
                    keep_trying = True
                    time.sleep(0.25)
                    break


def discover_cables(
    hw_server: ServerInfo,
    cs_server: Optional[ServerInfo],
    disable_core_scan: bool,
    timeout: int,
) -> QueryList[Cable]:
    """Given a hw_server and cs_server, scan for cables in hardware and return the list.

    Args:
        hw_server: hardware server connection
        cs_server: chipscope server connection (or None if no connection)
        disable_core_scan: Disable scanning debug cores in all devices (global to cable)
        timeout: cable device detection timeout in seconds

    Returns:
        New QueryList of cables
    """
    cables: QueryList[Cable] = QueryList()
    view = hw_server.get_view("jtag")
    for node in view.get_children():
        cables.append(_JtagCable(node, hw_server, cs_server, disable_core_scan, timeout))
    return cables



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from collections import namedtuple
from typing import List
from chipscopy.client.jtagdevice import JtagNode
from enum import Enum

JtagPadding = namedtuple("JtagPadding", ["size", "value"])


class JtagState(Enum):
    RESET = "RESET"
    IDLE = "IDLE"
    DRSELECT = "DRSELECT"
    DRCAPTURE = "DRCAPTURE"
    DRSHIFT = "DRSHIFT"
    DREXIT1 = "DREXIT1"
    DRPAUSE = "DRPAUSE"
    DREXIT2 = "DREXIT2"
    DRUPDATE = "DRUPDATE"
    IRSELECT = "IRSELECT"
    IRCAPTURE = "IRCAPTURE"
    IRSHIFT = "IRSHIFT"
    IREXIT1 = "IREXIT1"
    IRPAUSE = "IRPAUSE"
    IREXIT2 = "IREXIT2"
    IRUPDATE = "IRUPDATE"


class JtagSequence:
    _jtag_node: JtagNode = None
    _commands: list = None
    _data: bytes = None
    _ir_prefix = JtagPadding(size=0, value=0)
    _ir_postfix = JtagPadding(size=0, value=0)
    _dr_prefix = JtagPadding(size=0, value=0)
    _dr_postfix = JtagPadding(size=0, value=0)

    def __init__(self, jtag_node: JtagNode):
        self._jtag_node = jtag_node
        self._commands = list()
        self._data = bytearray()

    def __str__(self):
        return "Sequence commands = " + str(self._commands) + "\n" + "Data = " + str(self._data)

    def __repr__(self):
        return "Sequence commands = " + str(self._commands) + "\n" + "Data = " + str(self._data)

    def clear(self):
        """Remove all commands from JTAG sequence

        Args:
            None
        """
        self._commands.clear()
        self._data = bytearray()

    def set_ir_prefix(self, *, size: int, value: int):
        self._ir_prefix = JtagPadding(size=size, value=value)

    def set_ir_postfix(self, *, size: int, value: int):
        self._ir_postfix = JtagPadding(size=size, value=value)

    def set_dr_prefix(self, *, size: int, value: int):
        self._dr_prefix = JtagPadding(size=size, value=value)

    def set_dr_postfix(self, *, size: int, value: int):
        self._dr_postfix = JtagPadding(size=size, value=value)

    def shift(
        self,
        *,
        shift_type: str = "dr",
        data: int or bytearray = None,
        tdi: int = None,
        size,
        capture: bool = True,
        end_state: JtagState = JtagState.RESET,
    ):
        """Add JTAG sequence command to shift data in IR/DR SHIFT state

        Args:
            shift_type: JTAG state in which data to shift. Valid options: "dr" or "ir"
            data: Integer or bytearray data to shift
            tdi: TDI value to use for all clocks in shift state
            size: Size of data to be shifted in bits
            capture: Capture TDO data during shift and return from sequence run command
            end_state: JTAG state to enter after shift is complete. The default is RESET.

        TODO: Mask and compare options
        """

        # Check arguments
        if data is not None and tdi is not None:
            raise ValueError("Cannot use data and tdi shift arguments simultaneously.")

        if shift_type != "dr" and shift_type != "ir":
            raise ValueError('Valid shift_type values are "ir" and "dr".')

        # Create list with shift command arguments
        shift_command = ["shift", shift_type[0], capture, size]

        shift_options = dict()
        if end_state is not None:
            shift_options["state"] = end_state.name

        if data is not None:
            if isinstance(data, (bytes, bytearray)) is True:
                # Append data to object's member variable. Size should be ceil of size in bits / 8.
                self._data += data[: -(size // -8)]
            else:
                # If data is not in bytearray format convert it to bytes and append to object's member variable
                truncated_data = data & ((1 << size) - 1)
                self._data += truncated_data.to_bytes(-(size // -8), sys.byteorder)
        else:
            shift_options["value"] = tdi

        shift_command.append(shift_options)

        # Add command to list of commands
        self._commands.append(shift_command)

    def ir_shift(
        self,
        *,
        data=None,
        tdi: int = None,
        size,
        capture: bool = True,
        end_state: JtagState = JtagState.RESET,
    ):
        """Add JTAG sequence command to shift data in IR SHIFT state

        Args:
            data: Integer or bytearray data to shift
            tdi: TDI value to use for all clocks in shift state
            size: Size of data to be shifted in bits
            capture: Capture TDO data during shift and return from sequence run command
            end_state: JTAG state to enter after shift is complete. The default is RESET.
        """
        if self._ir_prefix.size != 0:
            self.shift(
                shift_type="ir",
                data=self._ir_prefix.value,
                size=self._ir_prefix.size,
                capture=False,
                end_state=JtagState.IRSHIFT,
            )

        self.shift(
            shift_type="ir",
            data=data,
            tdi=tdi,
            size=size,
            capture=capture,
            end_state=JtagState.IRSHIFT if self._ir_postfix.size != 0 else end_state,
        )

        if self._ir_postfix.size != 0:
            self.shift(
                shift_type="ir",
                data=self._ir_postfix.value,
                size=self._ir_postfix.size,
                capture=False,
                end_state=end_state,
            )

    def dr_shift(
        self,
        *,
        data=None,
        tdi: int = None,
        size,
        capture: bool = True,
        end_state: JtagState = JtagState.RESET,
    ):
        """Add JTAG sequence command to shift data in DR SHIFT state

        Args:
            data: Integer or bytearray data to shift
            tdi: TDI value to use for all clocks in shift state
            size: Size of data to be shifted in bits
            capture: Capture TDO data during shift and return from sequence run command
            end_state: JTAG state to enter after shift is complete. The default is RESET.
        """

        if self._dr_prefix.size != 0:
            self.shift(
                shift_type="dr",
                data=self._dr_prefix.value,
                size=self._dr_prefix.size,
                capture=False,
                end_state=JtagState.DRSHIFT,
            )

        self.shift(
            shift_type="dr",
            data=data,
            tdi=tdi,
            size=size,
            capture=capture,
            end_state=JtagState.DRSHIFT if self._dr_postfix.size != 0 else end_state,
        )

        if self._dr_postfix.size != 0:
            self.shift(
                shift_type="dr",
                data=self._dr_postfix.value,
                size=self._dr_postfix.size,
                capture=False,
                end_state=end_state,
            )

    def set_state(self, state: JtagState, count: int = 0):
        """Add JTAG sequence command to move JTAG state machine to a new state and then generate certain number of JTAG
        clocks

        Args:
            state: Move JTAG state machine to this state
            count: Number of JTAG clocks to generate after moving to specified JTAG state
        """
        state_command = ["state", state.name, count]
        self._commands.append(state_command)

    def run(self) -> List[bytearray]:
        """Run JTAG operations in sequence for target JTAG node. This method will return the result from IR/DR shift
        commands for which capture is true.

        Args:
            None

        TODO: Option to mask and compare data returned after running sequence
        """
        # Execute commands in JTAG sequence
        seq_result = self._jtag_node.sequence(self._commands, self._data)

        result = list()
        index = 0
        for command in self._commands:
            # Split combined sequence run result into a list of bytearrays based on length of shift commands in
            # JTAG sequence
            if command[0] == "shift" and command[2] is True:
                size_in_bytes = -(command[3] // -8)
                result.append(seq_result[index : index + size_in_bytes])
                index += size_in_bytes
        return result



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import ByteString, Dict, Any
from chipscopy.tcf.services import Service, DoneHWCommand, Token

NAME = "Recorder"
"""Recorder service name."""


class RecorderProxy(Service):
    """TCF Recorder service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def start_recorder(self, name: str, done: DoneHWCommand = None) -> Token:
        """
        Start TCF Recording

        :param name: Name of recording
        :param done: Callback with result and any error.
        """
        return self.send_xicom_command("startRecorder", (name,), done)

    def stop_recorder(self, name: str = "", done: DoneHWCommand = None) -> Token:
        """
        Stop TCF Recording

        :param name: Name of recording
        :param done: Callback with result and any error.
        """
        return self.send_xicom_command("stopRecorder", (name,), done)

    def start_playback(self, name: str, done: DoneHWCommand = None) -> Token:
        """
        Start TCF Recording

        :param name: Name of recording
        :param done: Callback with result and any error.
        """
        return self.send_xicom_command("startPlayback", (name,), done)

    def stop_playback(self, name: str = "", done: DoneHWCommand = None) -> Token:
        """
        Stop TCF Recording

        :param name: Name of recording
        :param done: Callback with result and any error.
        """
        return self.send_xicom_command("stopPlayback", (name,), done)
